from langchain_text_splitters import CharacterTextSplitter, RecursiveCharacterTextSplitter, TokenTextSplitter
from typing import Literal 
from langchain.schema import Document
from abc import ABC, abstractmethod

# HP: splitting_type, chunk_size, chunk_overlap
class BaseSplitter(ABC):
    """ Base Splitter Class provides 3 levels of splitting for our RAG pipeline
        based on this 
        https://github.com/FullStackRetrieval-com/RetrievalTutorials/blob/main/tutorials/LevelsOfTextSplitting/5_Levels_Of_Text_Splitting.ipynb

        You can extend it for your own needs
    """
    def __init__(self,
                 splitting_type: Literal["characters", "recursive_character", "tokens"],
                 chunk_size:int,
                 chunk_overlap:int=5,
                ) -> None:
                self.splitting_type = splitting_type
                self.chunk_size = chunk_size
                self.chunk_overlap = chunk_overlap

    @abstractmethod
    def split_documents(self,docs:list[Document]) -> list[Document]:
        if self.splitting_type == "characters":
            return self._split_characters(docs)
        elif self.splitting_type == "recursive_character":
            return self._split_recursive_character(docs)
        elif self.splitting_type == "tokens":
            return self._split_tokens(docs)
    
    def _split_characters(self,docs:list[Document]):
        text_splitter = CharacterTextSplitter(chunk_size = self.chunk_size, 
                                              chunk_overlap=self.chunk_overlap, 
                                              separator="\n\n",
                                              strip_whitespace=False)
        return text_splitter.split_documents(docs)

    def _split_recursive_character(self,docs:list[Document]):
        text_splitter = RecursiveCharacterTextSplitter(chunk_size = self.chunk_size, 
                                                        chunk_overlap=self.chunk_overlap, 
                                                        length_function=len)
        return text_splitter.split_documents(docs)
    
    def _split_tokens(self,docs:list[Document]):
        text_splitter = TokenTextSplitter(chunk_size = self.chunk_size, 
                                          chunk_overlap=self.chunk_overlap,
                                          encoding_name="cl100k_base")
        return text_splitter.split_documents(docs)

    def _split_sementic(self,docs:list[Document]):
        raise NotImplementedError()
    
    def _split_agentic(self,docs:list[Document]):
        raise NotImplementedError()

class Splitter(BaseSplitter):
    """ Splitter Class  for our RAG pipeline (Phase 2)
    """
    def __init__(self,
                 splitting_type: Literal["characters", "recursive_character", "tokens"],
                 chunk_size:int,
                 chunk_overlap:int=5,
                ) -> None:
        """ Load the parsed documents and split them

            Args: (those args represents hyperparameters in our RAG pipeline)
            splitting_type: Type of splitting criteria to use.
            chunk_size: Size of each chunk.
            chunk_overlap: Overlap between chunks.
        """
        super().__init__(splitting_type, chunk_size, chunk_overlap)
    
    def split_documents(self,docs:list[Document]) -> list[Document]:
        return super().split_documents(docs)

    


from typing import Optional, Any, Callable
from typing_extensions import Annotated, Doc, TypeAlias, Literal
from concurrent.futures import Future, as_completed, Executor
from rag_opt.dataset import TrainDataset, EvaluationDataset
from rag_opt.search_space import RAGSearchSpace
from rag_opt._utils import get_shared_executor
from rag_opt._sampler import SamplerType
from rag_opt._config import RAGConfig
from rag_opt.rag import RAGWorkflow
from langchain.chat_models.base import BaseChatModel
from langchain.schema.embeddings import Embeddings
from langchain.schema import Document
from threading import Lock
from loguru import logger
import torch
import os

ComponentType: TypeAlias = Literal["llms", "embeddings", "vector_stores", "rerankers"]


class RAGPipelineManager:
    """
    RAG Pipeline Manager with Gateway API support.
    """

    def __init__(
        self,
        search_space: Annotated[RAGSearchSpace, Doc("RAG search space to optimize")],
        *,
        max_workers: Annotated[int, Doc("Maximum workers for parallel operations")] = 5,
        eager_load: Annotated[bool, Doc("Load all components immediately")] = False,
        verbose: Annotated[bool, Doc("Enable verbose logging")] = False,
        executor: Annotated[Optional[Executor], Doc("Thread pool executor")] = None,
        api_key: Annotated[Optional[str], Doc("Universal API key (for gateway or providers)")] = None,
        **kwargs
    ):
        self._search_space = search_space
        self.max_workers = max_workers
        self.eager_load = eager_load
        self._verbose = verbose
        self.executor = executor or get_shared_executor(max_workers)
        
        # Universal API key handling
        self.api_key = api_key or os.getenv('AI_GATEWAY_API_KEY')
        
        # Component registry and lock
        self._registry: dict[str, Any] = {}
        self._lock = Lock()
        self._init_kwargs = kwargs
        
        # Load components if eager loading is enabled
        if eager_load:
            logger.debug("RAGPipelineManager: Eagerly loading all components")
            self._load_all_components_parallel()

    @property
    def search_space(self) -> RAGSearchSpace:
        """Get the search space."""
        return self._search_space

    @property
    def verbose(self) -> bool:
        """Get verbose flag."""
        return self._verbose

    def _build_cache_key(self, component_type: ComponentType, **kwargs) -> str:
        """Generate unique cache key for component type and parameters."""
        if component_type not in {"llms", "embeddings", "vector_stores", "rerankers"}:
            raise ValueError(f"Invalid component type: {component_type}")
        
        parts = [component_type]
        for key, value in sorted(kwargs.items()):
            if value is not None:
                parts.append(f"{key}={value}")
        return "|".join(parts)

    def _get_or_create_component(
        self,
        cache_key: str,
        factory_func: Callable[..., Any],
        *args,
        **kwargs
    ) -> Any:
        """Thread-safe component creation with caching."""
        # Fast path: check without lock
        if cache_key in self._registry:
            return self._registry[cache_key]
        
        with self._lock:
            if cache_key not in self._registry:
                try:
                    self._registry[cache_key] = factory_func(*args, **kwargs)
                    if self._verbose:
                        logger.debug(f"Created and cached: {cache_key}")
                except Exception as e:
                    logger.error(f"Failed to create component {cache_key}: {e}")
                    raise
            return self._registry[cache_key]

    def _is_gateway_model(self, model: str) -> bool:
        """Check if model string is in gateway format (provider/model)."""
        return '/' in model

    def get_llm(self, model: str, provider: str, api_key: Optional[str] = None) -> BaseChatModel:
        """Get or create LLM instance - supports both gateway and standard formats."""
        # Use provided API key, fallback to instance key
        effective_api_key = api_key or self.api_key
        
        # Gateway format: provider/model
        if self._is_gateway_model(model) or provider == 'gateway':
            gateway_model = model if '/' in model else f"{provider}/{model}"
            cache_key = self._build_cache_key("llms", model=gateway_model)
            
            from rag_opt.llm import init_chat_model
            return self._get_or_create_component(
                cache_key,
                init_chat_model,
                model=gateway_model,
                api_key=effective_api_key
            )
        else:
            # Standard provider
            cache_key = self._build_cache_key("llms", model=model, provider=provider)
            
            from rag_opt import init_chat_model
            return self._get_or_create_component(
                cache_key,
                init_chat_model,
                model=model,
                model_provider=provider,
                api_key=effective_api_key
            )

    def get_embeddings(self, provider: str, model: str, api_key: Optional[str] = None) -> Embeddings:
        """Get or create embeddings instance - supports both gateway and standard formats."""
        # Use provided API key, fallback to instance key
        effective_api_key = api_key or self.api_key

        # Gateway format: provider/model
        if self._is_gateway_model(model) or provider == 'gateway':
            gateway_model = model if '/' in model else f"{provider}/{model}"
            cache_key = self._build_cache_key("embeddings", model=gateway_model)
            
            from rag_opt.llm import init_embeddings
            return self._get_or_create_component(
                cache_key,
                init_embeddings,
                model=gateway_model,
                api_key=effective_api_key
            )
        else:
            # Standard provider
            cache_key = self._build_cache_key("embeddings", provider=provider, model=model)
            
            from rag_opt import init_embeddings
            return self._get_or_create_component(
                cache_key,
                init_embeddings,
                model_provider=provider,
                model=model,
                api_key=effective_api_key
            )

    def get_reranker(self, model: str, provider: str, api_key: Optional[str] = None) -> Optional[Any]:
        """Get or create reranker instance."""
        if not provider or not model:
            return None
        
        effective_api_key = api_key or self.api_key
        cache_key = self._build_cache_key("rerankers", model=model, provider=provider)
        
        from rag_opt import init_reranker
        return self._get_or_create_component(
            cache_key,
            init_reranker,
            model=model,
            provider=provider,
            api_key=effective_api_key
        )

    def get_vector_store(
        self,
        provider: str,
        embeddings: Embeddings,
        documents: list[Document],
        index_name: Optional[str] = None,
        api_key: Optional[str] = None,
        initialize: bool = False
    ) -> Any:
        """Get or create vector store instance with smart caching."""
        from rag_opt import init_vectorstore
        
        effective_api_key = api_key or self.api_key
        
        # Cache based on provider and embedding type only
        cache_key = self._build_cache_key(
            "vector_stores",
            provider=provider,
            embedding_class=embeddings.__class__.__name__
        )
        
        return self._get_or_create_component(
            cache_key,
            init_vectorstore,
            provider,
            embeddings,
            documents=documents,
            index_name=index_name,
            api_key=effective_api_key,
            initialize=initialize
        )

    def _load_component_batch(
        self,
        component_type: ComponentType,
        configs: list[dict],
        getter_func: Callable
    ) -> None:
        """Load a batch of components in parallel."""
        futures = []
        
        for config in configs:
            future = self.executor.submit(getter_func, **config)
            futures.append(future)
        
        # Wait for completion
        for future in as_completed(futures):
            try:
                future.result()
            except Exception as e:
                logger.warning(f"Component load failed: {e}")

    def _load_all_components_parallel(self) -> None:
        """Load all components from search space in parallel."""
        load_tasks = []

        # Load LLMs
        llm_configs = []
        if self._search_space.llm and self._search_space.llm.choices:
            for provider_name, config in self._search_space.llm.choices.items():
                # Handle gateway format (list of model strings)
                if provider_name == 'gateway' and isinstance(config, list):
                    for model in config:
                        llm_configs.append({
                            "model": model,
                            "provider": provider_name,
                            "api_key": self.api_key
                        })
                # Handle standard provider format (config object)
                elif hasattr(config, 'models'):
                    for model in config.models:
                        llm_configs.append({
                            "model": model,
                            "provider": provider_name,
                            "api_key": config.api_key or self.api_key
                        })
        if llm_configs:
            load_tasks.append(("llms", llm_configs, self.get_llm))

        # Load Embeddings
        emb_configs = []
        if self._search_space.embedding and self._search_space.embedding.choices:
            for provider_name, config in self._search_space.embedding.choices.items():
                # Handle gateway format (list of model strings)
                if provider_name == 'gateway' and isinstance(config, list):
                    for model in config:
                        emb_configs.append({
                            "provider": provider_name,
                            "model": model,
                            "api_key": self.api_key
                        })
                # Handle standard provider format (config object)
                elif hasattr(config, 'models'):
                    for model in config.models:
                        emb_configs.append({
                            "provider": provider_name,
                            "model": model,
                            "api_key": config.api_key or self.api_key
                        })
        if emb_configs:
            load_tasks.append(("embeddings", emb_configs, self.get_embeddings))

        # Load Rerankers
        reranker_configs = []
        if self._search_space.reranker and self._search_space.reranker.choices:
            for provider_name, config in self._search_space.reranker.choices.items():
                # Handle gateway format (list of model strings)
                if provider_name == 'gateway' and isinstance(config, list):
                    for model in config:
                        if model is not None:
                            reranker_configs.append({
                                "model": model,
                                "provider": provider_name,
                                "api_key": self.api_key
                            })
                # Handle standard provider format (config object)
                elif hasattr(config, 'models'):
                    for model in config.models:
                        if model is not None: 
                            reranker_configs.append({
                                "model": model,
                                "provider": provider_name,
                                "api_key": config.api_key or self.api_key
                            })
        if reranker_configs:
            load_tasks.append(("rerankers", reranker_configs, self.get_reranker))

        # Execute loading
        for component_type, configs, getter_func in load_tasks:
            self._load_component_batch(component_type, configs, getter_func)

    def initiate_llm(self, model_name: Optional[str] = None) -> BaseChatModel:
        """Get LLM for evaluation - supports gateway format (provider/model)."""
        # Handle gateway format
        if model_name and '/' in model_name:
            provider, model = model_name.split('/', 1)
            return self.get_llm(model=model, provider=provider, api_key=self.api_key)
        
        # Try to get cached LLM by name
        if model_name:
            for key, component in self._registry.items():
                if "llms" in key and model_name in key:
                    return component
        
        # Get first available from search space
        if hasattr(self._search_space, 'llm') and self._search_space.llm.choices:
            for provider_name, config in self._search_space.llm.choices.items():
                if config.models:
                    return self.get_llm(
                        model=config.models[0],
                        provider=provider_name,
                        api_key=config.api_key or self.api_key
                    )
        
        raise ValueError("No LLMs available in search space")

    def initiate_embedding(self, model_name: Optional[str] = None) -> Embeddings:
        """Get embedding for evaluation - supports gateway format (provider/model)."""
        # Handle gateway format
        if model_name and '/' in model_name:
            provider, model = model_name.split('/', 1)
            return self.get_embeddings(provider=provider, model=model, api_key=self.api_key)
        
        # Try to get cached embedding by name
        if model_name:
            for key, component in self._registry.items():
                if "embeddings" in key and model_name in key:
                    return component
        
        # Get first available from search space
        if hasattr(self._search_space, 'embedding') and self._search_space.embedding.choices:
            for provider_name, config in self._search_space.embedding.choices.items():
                if config.models:
                    return self.get_embeddings(
                        provider=provider_name,
                        model=config.models[0],
                        api_key=config.api_key or self.api_key
                    )
        
        raise ValueError("No embeddings available in search space")

    def create_rag_instance(
        self,
        config: RAGConfig,
        documents: Optional[list[Document]] = None,
        retrieval_config: Optional[dict] = None,
        initialize: bool = False
    ) -> RAGWorkflow:
        """Create RAGWorkflow instance from configuration using cached components."""
        llm = self.get_llm(
            model=config.llm.model,
            provider=config.llm.provider,
            api_key=config.llm.api_key or self.api_key
        )
        
        embeddings = self.get_embeddings(
            provider=config.embedding.provider,
            model=config.embedding.model,
            api_key=config.embedding.api_key or self.api_key
        )
        
        reranker = None
        if config.reranker:
            reranker = self.get_reranker(
                model=config.reranker.model,
                provider=config.reranker.provider,
                api_key=config.reranker.api_key or self.api_key
            )
        
        vector_store = self.get_vector_store(
            provider=config.vector_store.provider,
            embeddings=embeddings,
            documents=documents or [],
            index_name=config.vector_store.index_name,
            api_key=config.vector_store.api_key or self.api_key,
            initialize=initialize
        )
        
        return RAGWorkflow(
            embeddings=embeddings,
            vector_store=vector_store,
            llm=llm,
            reranker=reranker,
            llm_provider_name=config.llm.provider,
            llm_model_name=config.llm.model,
            embedding_provider_name=config.embedding.provider,
            embedding_model_name=config.embedding.model,
            reranker_provider_name=config.reranker.provider if config.reranker else None,
            reranker_model_name=config.reranker.model if config.reranker else None,
            vector_store_provider_name=config.vector_store.provider,
            retrieval_config=retrieval_config or {"search_type": config.search_type, "k": config.k},
            corpus_documents=documents,
            max_workers=self.max_workers,
            executor=self.executor
        )

    def generate_initial_data(
        self,
        train_data: TrainDataset,
        n_samples: int = 20,
        sampler_type: SamplerType = SamplerType.SOBOL,
        **kwargs
    ) -> tuple[list[RAGConfig], list[EvaluationDataset]]:
        """Generate initial data with parallel execution."""
        rag_configs = self._search_space.sample(n_samples=n_samples, sampler_type=sampler_type)
        
        documents = train_data.to_langchain_docs()
        configs: list[RAGConfig] = []
        datasets: list[EvaluationDataset] = []
        future_map: dict[Future[EvaluationDataset], RAGConfig] = {}
        
        # Submit all tasks in parallel
        for rag_config in rag_configs:
            rag = self.create_rag_instance(rag_config, documents=documents, initialize=True, **kwargs)
            future = self.executor.submit(
                rag.get_batch_answers,
                dataset=train_data,
                **rag_config.to_dict()
            )
            future_map[future] = rag_config
        
        # Collect results as they complete
        for future in as_completed(future_map):
            rag_config = future_map[future]
            try:
                dataset = future.result(timeout=300)
                if len(dataset.items) > 0:
                    datasets.append(dataset)
                    configs.append(rag_config)
                else:
                    logger.warning(f"Empty dataset for config {rag_config}")
            except Exception as e:
                logger.error(f"Error processing config {rag_config}: {e}")
        
        return configs, datasets

    def create_rag_instance_by_sample(
        self,
        sampler_type: SamplerType = SamplerType.SOBOL,
        documents: Optional[list[Document]] = None,
        retrieval_config: Optional[dict] = None
    ) -> RAGWorkflow:
        """Create RAGWorkflow instance from a sampled search space config."""
        sample = self._search_space.sample(n_samples=1, sampler_type=sampler_type)
        if not sample:
            logger.error("No sample found in search space")
            raise ValueError("No sample found in search space")
        return self.create_rag_instance(sample[0], documents=documents, retrieval_config=retrieval_config)

    def generate_evaluation_data(
        self, 
        config: RAGConfig, 
        train_data: TrainDataset, 
        **kwargs
    ) -> EvaluationDataset:
        """Generate evaluation dataset from a sampled search space config."""
        documents = train_data.to_langchain_docs()
        rag = self.create_rag_instance(config, documents=documents, **kwargs)
        return rag.get_batch_answers(dataset=train_data, **kwargs)

    def clear_cache(self) -> None:
        """Clear all cached components."""
        with self._lock:
            self._registry.clear()
            logger.info("Component cache cleared")

    def sample(self, n_samples: int = 1, **kwargs) -> list[RAGConfig]:
        """Sample configurations from the search space."""
        return self._search_space.sample(n_samples, **kwargs)

    def get_problem_bounds(self) -> torch.Tensor:
        """Get parameter bounds from search space."""
        return self._search_space.get_parameter_bounds()

    def decode_sample_to_rag_config(self, sample: torch.Tensor) -> RAGConfig:
        """Decode a sample generated tensor to RAGConfig."""
        return self._search_space.decode_sample_to_rag_config(sample)

    def encode_rag_config_to_tensor(self, config: RAGConfig) -> torch.Tensor:
        """Encode a RAGConfig to a tensor."""
        return self._search_space.config_to_tensor(config)

    @classmethod
    def from_search_space(
        cls,
        search_space: RAGSearchSpace,
        max_workers: int = 4,
        eager_load: bool = False,
        **kwargs
    ) -> "RAGPipelineManager":
        """Factory method to create manager from search space."""
        return cls(
            search_space=search_space,
            max_workers=max_workers,
            eager_load=eager_load,
            **kwargs
        )


from __future__ import annotations
from typing import Optional, Any, Annotated, TypeVar, Generic, Literal
from pydantic import BaseModel, Field
from langchain.schema import Document



QuestionDifficulty = Literal["easy", "medium", "hard"]

class GroundTruth(BaseModel):
    """ Generated from the train dataset item and used later for evaluation metric """
    answer: str
    contexts: list[str]

class ComponentUsage(BaseModel):
    """Base class for component-level metrics (cost, latency, tokens, etc.)"""
    llm: Annotated[float, Field(description="usage info of generation for one query like cost , latency, ...")]
    embedding: Annotated[float, Field(description="usage info of embedding for one query like cost , latency, ...")]
    vectorstore: Annotated[float, Field(description="usage info of vectorstore for one query like cost , latency, ...")]
    reranker: Annotated[float, Field(description="usage info of reranker for one query like cost , latency, ...")]

    @property
    def total(self) -> float:
        return self.llm + self.embedding + self.vectorstore + self.reranker


class TrainDatasetItem(BaseModel):
    question: str 
    answer: str 
    contexts: list[str] 
    difficulty: Annotated[Optional[QuestionDifficulty], Field(description="The difficulty level of the question", default=None)]
    def to_ground_truth(self) -> GroundTruth:
        """Convert training item to ground truth for evaluation"""
        return GroundTruth(answer=self.answer, contexts=self.contexts)
    
class EvaluationDatasetItem(BaseModel):
    question: str
    answer: str
    contexts: list[str]
    ground_truth: Annotated[Optional[GroundTruth], Field(description="The expected answer and contexts for the question")] = None
    metadata: Annotated[Optional[dict[str, Any]], Field(default_factory=dict, description="Generated information from callbacks for evaluation metrics")] 
    
    @property
    def cost(self) -> ComponentUsage:
        """ Cost metrics for one query """
        if "cost" in self.metadata:
            if isinstance(self.metadata["cost"], dict):
                return ComponentUsage(**self.metadata["cost"])
            return self.metadata["cost"]
        return ComponentUsage(llm=0, embedding=0, vectorstore=0, reranker=0)
    
    @property
    def latency(self) -> ComponentUsage:
        """ Latency metrics for one query """
        if "latency" in self.metadata:
            if isinstance(self.metadata["latency"], dict):
                return ComponentUsage(**self.metadata["latency"])
            return self.metadata["latency"]
        return ComponentUsage(llm=0, embedding=0, vectorstore=0, reranker=0)

    def set_cost(self, cost: ComponentUsage | dict) -> None:
        """Store cost metrics in metadata"""
        self.metadata["cost"] = cost.model_dump() if isinstance(cost, ComponentUsage) else cost
    
    def set_latency(self, latency: ComponentUsage | dict) -> None:
        """Store latency metrics in metadata"""
        self.metadata["latency"] = latency.model_dump() if isinstance(latency, ComponentUsage) else latency

T = TypeVar('T', bound=TrainDatasetItem|EvaluationDatasetItem)

class DatbasetMixin(BaseModel, Generic[T]):
    items: list[T]

    def to_langchain_docs(self) -> list[Document]:
        """Convert dataset items to langchain documents"""
        if self.items and isinstance(self.items[0], TrainDatasetItem): # train has no metadata 
            return [Document(page_content=f"Question: {d.question}\nAnswer: {d.answer}\nContexts: {d.contexts}") for d in self.items]
        
        return [Document(page_content=f"Question: {d.question}\nAnswer: {d.answer}", metadata=d.metadata) for d in self.items]

    def to_ground_truth(self) -> list[GroundTruth]:
        return [item.to_ground_truth() for item in self.items]
    
    def to_json(self, path: str = "./rag_dataset.json") -> None:
        with open(path, 'w') as f:
            f.write(self.model_dump_json(indent=2))
    
    @classmethod
    def from_json(cls, path: str) -> DatbasetMixin[T]:
        with open(path, 'r') as f:
            return cls.model_validate_json(f.read())
    
    def __len__(self):
        return len(self.items)

class TrainDataset(DatbasetMixin[TrainDatasetItem]):
    """ dataset used in querying process"""
    items: list[TrainDatasetItem]
    

class EvaluationDataset(DatbasetMixin[EvaluationDatasetItem]):
    items: list[EvaluationDatasetItem]
    
    




from rag_opt.eval.metrics.base import BaseMetric, MetricResult, MetricCategory
from rag_opt.eval.metrics.retrieval import ContextPrecision
from rag_opt.eval.metrics.full import CostMetric, LatencyMetric
from rag_opt.eval.metrics.generation import ResponseRelevancy, SafetyMetric, AlignmentMetric
from rag_opt.eval.metrics.retrieval import ContextPrecision, ContextRecall, MRR, NDCG
from rag_opt.llm import RAGLLM



__all__ = [
    "BaseMetric",
    "MetricResult",
    "MetricCategory",
    "ContextPrecision",
    "CostMetric",
    "LatencyMetric",
    "ResponseRelevancy",
    "SafetyMetric",
    "AlignmentMetric",
    "ContextPrecision",
    "ContextRecall",
    "MRR",
    "NDCG"
]


from rag_opt.dataset import TrainDataset, TrainDatasetItem, QuestionDifficulty
from rag_opt._prompts import RAG_DATASET_GENERATION_PROMPT
from langchain.schema import Document, BaseMessage
from langchain_core.prompts import PromptTemplate
from langchain.schema import BasePromptTemplate
from typing_extensions import Doc, Annotated
from rag_opt.rag import Parser
from rag_opt.llm import RAGLLM
from typing import Optional
import rag_opt._utils as _utils
from loguru import logger 
import asyncio
import random
import json


DIFFICULTY_CONFIGS = {
    "easy": {
        "description": "Generate basic recall questions that can be answered directly from the text",
        "sample_questions": ["What is X?", "Where does Y happen?", "When did Z occur?"]
    },
    "medium": {
        "description": "Generate analytical questions requiring understanding and synthesis",
        "sample_questions": ["How does X relate to Y?", "What are the implications of Z?", "Compare X and Y"]
    },
    "hard": {
        "description": "Generate complex questions requiring deep reasoning and inference",
        "sample_questions": ["Why might X lead to Y?", "What would happen if Z changed?", "Evaluate the effectiveness of X"]
    }
}

class DatasetGenerator:
    """ Prepreare list of Training dataset to be used as ground truth for evaluation process"""
    _prompt: BasePromptTemplate = None 
    def __init__(self, 
                 llm: Annotated[RAGLLM, Doc("the llm to be used in the dataset generation process")],
                 dataset_path: Annotated[str, Doc("the path to the dataset folder or file to be used by parser")] = "./data",
                 *,
                 custom_prompt: Annotated[str, Doc("the custom prompt template to be used for question-answer generation")] = None,
                 parser: Annotated[Parser, Doc("the dataset loader to load and parse dataset document from a specific folder or file")] = None,
                 ):
        
        self.llm = llm

        if not parser and not dataset_path:
            logger.error("You must provide either parser or dataset_path")
            raise ValueError("Must provide either parser or dataset_path")

        # parser
        if dataset_path and not parser:
            logger.info(f"Loading documents from {dataset_path}")
            parser= Parser(path=dataset_path)
        
        self.parser = parser
        self.dataset_path = dataset_path
        
        if custom_prompt:
            _utils.validate_prompt(custom_prompt, RAG_DATASET_GENERATION_PROMPT, raise_error=True)

        self._prompt = PromptTemplate(
            template=custom_prompt or RAG_DATASET_GENERATION_PROMPT,
            input_variables=["contexts", "difficulty_instruction"]
        )

    def _load_documents(self) -> list[Document]:
        """Load documents using parser if dataset_path is provided"""
        if not self.parser or not self.dataset_path:
            return []
        try:
            return self.parser.load_docs()
        except Exception as e:
            logger.error(f"Failed to load documents from {self.dataset_path}: {e}")
            return []

    def _get_difficulty_instruction(self, difficulty: QuestionDifficulty) -> str:
        """Get difficulty-specific instruction for prompt"""
        config = DIFFICULTY_CONFIGS[difficulty]
        return f"{config['description']}. Examples: {', '.join(config['sample_questions'])}"
    
    def _create_batch_prompts_with_contexts(self, contexts_batch: list[str], difficulties: list[QuestionDifficulty]) -> list[tuple[str, list[str]]]:
        """Create prompts for a batch of contexts with varying difficulties"""
        prompts_contexts = []
        for contexts_subset, difficulty in zip(self._chunk_contexts(contexts_batch, len(difficulties)), difficulties):
            difficulty_instruction = self._get_difficulty_instruction(difficulty)
            contexts= random.sample(contexts_subset, min(3,len(contexts_subset)))
            prompt = self._prompt.format(
                contexts=contexts,
                difficulty_instruction=difficulty_instruction
            )
            prompts_contexts.append((prompt, contexts))
        return prompts_contexts
    
    def _chunk_contexts(self, contexts: list[str], num_chunks: int) -> list[list[str]]:
        """Split contexts into chunks"""
        chunk_size = max(1, len(contexts) // max(1, num_chunks))
        return [contexts[i:i + chunk_size] for i in range(0, len(contexts), chunk_size)]
    
    def _parse_llm_responses(self, 
                             responses: list[BaseMessage], 
                             difficulties: list[QuestionDifficulty],
                             used_contexts: list[list[str]]) -> list[TrainDatasetItem]:
        """Parse LLM responses into dataset items"""
        items = []
        for response, difficulty, contexts in zip(responses, difficulties, used_contexts):
            try:
                data = json.loads(response.content if hasattr(response, 'content') else str(response))
                if isinstance(data, list):
                    for qa in data:
                        items.append(self._create_dataset_item(qa, difficulty,contexts=contexts))
                else:
                    items.append(self._create_dataset_item(data, difficulty,contexts=contexts))
                    
            except json.JSONDecodeError:
                fallback_item = self._parse_fallback_response(str(response), contexts, difficulty)
                if fallback_item:
                    items.append(fallback_item)
        return items

    def _create_dataset_item(self, qa_data: dict, 
                             difficulty: QuestionDifficulty,
                             contexts: list[str]) -> TrainDatasetItem:
        """Create TrainDatasetItem from parsed data"""
        return TrainDatasetItem(
            question=qa_data.get('question', ''),
            answer=qa_data.get('answer', ''),
            contexts=contexts,
            difficulty=difficulty
        )
        
    def _generate_batch_contexts_and_size(self,
                                 source_texts: Optional[list[str]] = None,
                                 source_docs: Optional[list[Document]] = None,
                                 n: int = 10,
                                 batch_size: Optional[int] = None,
                                 ) -> tuple[list[list[str]], int]:
        if not source_docs and not source_texts:
            # Try to load from parser if available
            if self.parser and self.dataset_path:
                logger.info(f"Loading documents from {self.dataset_path}")
                source_docs = self._load_documents()
                if not source_docs:
                    logger.error("Failed to load documents from parser")
                    raise ValueError("Failed to load documents from parser")
            else:
                logger.error("You have to provide at least one of source_texts or source_docs")
                raise ValueError("You have to provide at least one of source_texts or source_docs")

        if not source_texts:
            source_texts = [doc.page_content for doc in source_docs]
        
        source_texts = [text.strip() for text in source_texts if text.strip()]
        if not source_texts:
            raise ValueError("No valid source texts found")

        batch_size = batch_size or max(min(n // 4, 1), 5, len(source_texts))
        logger.info(f"Generating {n} QA pairs with batch size {batch_size}")
        
        _batch_context_length = len(source_texts) // batch_size
        batches_contexts = [source_texts[i:i + _batch_context_length] for i in range(0, len(source_texts), _batch_context_length)]
        return batches_contexts, batch_size

    def _generate_difficulty_distribution(self, n: int) -> list[QuestionDifficulty]:
        """Generate balanced difficulty distribution"""
        # Default distribution: 30% easy, 50% medium, 20% hard
        easy_count = max(1, int(n * 0.3))
        hard_count = max(1, int(n * 0.2))
        medium_count = max(1,n - easy_count - hard_count)
        difficulties = (["easy"] * easy_count + 
                       ["medium"] * medium_count + 
                       ["hard"] * hard_count)
        
        random.shuffle(difficulties)
        return difficulties
    def _prepare_source_data(self, source_texts: Optional[list[str]], 
                           source_docs: Optional[list[Document]]) -> list[str]:
        """Prepare and validate source data"""
        if not source_docs and not source_texts:
            if self.parser and self.dataset_path:
                logger.info(f"Loading documents from {self.dataset_path}")
                source_docs = self._load_documents()
                if not source_docs:
                    raise ValueError("Failed to load documents from parser")
            else:
                raise ValueError("Must provide source_texts, source_docs, or dataset_path")

        # Convert docs to texts if needed
        if not source_texts and source_docs:
            source_texts = [doc.page_content for doc in source_docs]
        
        # Filter empty texts
        return [text.strip() for text in source_texts if text.strip()]
    def _parse_fallback_response(self, response: str, contexts: list[str], 
                               difficulty: QuestionDifficulty) -> TrainDatasetItem | None:
        """Fallback parser for non-JSON responses"""
        try:
            lines = [line.strip() for line in response.split('\n') if line.strip()]
            question = answer = ""
            
            for line in lines:
                lower_line = line.lower()
                if lower_line.startswith(('question:', 'q:')):
                    question = line.split(':', 1)[1].strip()
                elif lower_line.startswith(('answer:', 'a:')):
                    answer = line.split(':', 1)[1].strip()
            
            if question and answer:
                return TrainDatasetItem(
                    question=question,
                    answer=answer,
                    contexts=contexts,
                    difficulty=difficulty
                )
        except Exception as e:
            logger.error(f"Fallback parsing failed: {e}")
        
        return None
    
    def generate(self, 
             n: int = 10,
             *,
             source_texts: Optional[list[str]] = None,
             source_docs: Optional[list[Document]] = None,
             batch_size: Optional[int] = None) -> TrainDataset:
        """Generate synthetic RAG dataset with question difficulty variance
        
        Args:
            source_texts: Source texts for question generation
            source_docs: Source documents for question generation  
            n: Number of questions to generate
            batch_size: Batch size for generation (default: adaptive)
            
        Returns:
            TrainDataset: Generated dataset with varied question difficulties
        """
        texts = self._prepare_source_data(source_texts, source_docs)
        
        if len(texts) < n:
            logger.warning(f"Not enough source texts to generate {n} items (got {len(texts)}) - using all available texts {len(texts)}")
            n = len(texts)
        
        texts = texts[:n]
        
        # Determine batch size
        batch_size = batch_size or min(max(1, n // 4), 5)
        difficulties = self._generate_difficulty_distribution(n)
        
        # Split texts and difficulties into batches of batch_size
        text_batches = self._chunk_contexts(texts, batch_size)
        difficulty_batches = self._chunk_contexts(difficulties, batch_size)
        
        all_items = []
        
        logger.info(f"Generating {len(text_batches)} batches of up to {batch_size} QA pairs")
        
        try:
            for i, (text_batch, diff_batch) in enumerate(zip(text_batches, difficulty_batches)):
                logger.info(f"Generating batch {i+1}/{len(text_batches)}")
                
                prompts_contexts = self._create_batch_prompts_with_contexts(text_batch, diff_batch)
                prompts = [prompt for prompt, _ in prompts_contexts]
                contexts = [context for _, context in prompts_contexts]
                responses = self.llm.batch(prompts)
                
                batch_items = self._parse_llm_responses(responses, diff_batch, used_contexts=contexts)
                all_items.extend(batch_items)
                
        except KeyboardInterrupt:
            logger.warning("Generation interrupted by user")
        except Exception as e:
            logger.error(f"Error during generation: {e}")
        
        # Ensure exactly n items
        if len(all_items) > n:
            all_items = all_items[:n]
        
        logger.success(f"Generated {len(all_items)} QA pairs")
        return TrainDataset(items=all_items)

    
    async def agenerate(self,
                        n: int = 10,
                        *,
                        source_texts: Optional[list[str]] = None,
                        source_docs: Optional[list[Document]] = None,
                        batch_size: Optional[int] = None,
                        **kwargs):
        """ async generate datasets"""
        return await asyncio.to_thread(self.generate,n=n, source_texts=source_texts, source_docs=source_docs, batch_size=batch_size, **kwargs)


from langchain_unstructured import UnstructuredLoader
from langchain_community.document_loaders import (
                                            CSVLoader,
                                            UnstructuredExcelLoader,
                                            PyPDFLoader,
                                            # UnstructuredFileLoader,
                                            TextLoader,
                                            DirectoryLoader,
                                            DataFrameLoader
                                        )
from langchain_community.document_loaders.base import BaseLoader
from langchain.schema import Document
from typing import TYPE_CHECKING,  Any, Optional,  Iterator
import tempfile
import logging
import shutil
import os
import pandas as pd
from pathlib import Path
from loguru import logger

if TYPE_CHECKING:
    from pandas import DataFrame

logging.getLogger("pdfminer").setLevel(logging.ERROR)

class CustomFileLoader(BaseLoader):
    """Custom loader that uses appropriate loader based on file type."""
    
    def __init__(self, file_path: str, **kwargs):
        self.file_path = file_path
        self.kwargs = kwargs
        
    def lazy_load(self) -> Iterator[Document]:
        file_path = Path(self.file_path)
        file_extension = file_path.suffix.lower()
        
        if file_extension == '.csv':
            loader = CSVLoader(
                file_path=str(file_path),
                **self.kwargs
            )
        elif file_extension == '.xlsx' or file_extension == '.xls':
            loader = UnstructuredExcelLoader(
                file_path=str(file_path),
                mode="single",
                **self.kwargs
            )
        elif file_extension == '.pdf':
            loader = PyPDFLoader(
                str(file_path),
                **self.kwargs
            )
        elif file_extension == '.txt':
            loader = TextLoader(
                str(file_path),
                **self.kwargs
            )
        else:
            loader = UnstructuredLoader(
                str(file_path),
                **self.kwargs
            )
        try:
            for doc in loader.lazy_load():
                yield doc
        except NotImplementedError:
            # Fallback if lazy_load not implemented
            for doc in loader.load():
                yield doc
            
class Parser:
    """Enhanced Parser with custom loader support.
    This represents Phase 1 in RAG pipeline.
    """
    
    def __init__(self, 
                 path: str,
                 glob: str = "**/[!.]*",
                 include_sub_dir: bool = False,
                 silent_errors: bool = True,
                 use_multithreading: bool = True,
                 max_concurrency: int = 4,
                 csv_loader_kwargs: Optional[dict[str, Any]] = None,
                 excel_loader_kwargs: Optional[dict[str, Any]] = None,
                 **kwargs):
        """Initialize parser with path to data and custom loader options.
        
        Args: 
            path: Path to directory or file.
            glob: A glob pattern or list of glob patterns to use to find files.
            csv_loader_kwargs: Keyword arguments for CSVDocumentLoader
            excel_loader_kwargs: Keyword arguments for ExcelDocumentLoader
            **kwargs: Additional arguments for DirectoryLoader.
        """
        self._tmp_dir: str = None
        self.csv_loader_kwargs = csv_loader_kwargs or {}
        self.excel_loader_kwargs = excel_loader_kwargs or {}
        
        # Check if it is a file, create tmp dir and load 
        if not os.path.isdir(path):
            path = self._create_tmp_dir(path)
            self._tmp_dir = path
            
     
        self.loader = DirectoryLoader(
            path,
            glob=glob,
            silent_errors=silent_errors,
            show_progress=True,
            recursive=include_sub_dir,
            use_multithreading=use_multithreading,
            max_concurrency=max_concurrency,
            loader_cls=CustomFileLoader,
            **kwargs
        )
    
    def _create_tmp_dir(self, file_path: str) -> str:
        """Create temporary directory in case user provides a file not directory."""
        tmp_dir = tempfile.mkdtemp()  
        basename = os.path.basename(file_path)
        tmp_file_path = os.path.join(tmp_dir, basename)
        shutil.copy(file_path, tmp_file_path)
        return tmp_dir
    
    def load_text(self) -> list[str]:
        """Load the data folder files into list of strings."""
        docs = self.load_docs()
        logger.debug(f"Retrieved Contexts: {docs[0]}")
        raw_texts = [doc.page_content for doc in docs]
        return raw_texts
    
    def load_docs(self) -> list[Document]:
        """Load the data folder files into langchain documents."""
        docs = self.loader.load()
        if self._tmp_dir:
            shutil.rmtree(self._tmp_dir)
            self._tmp_dir = None
        return docs 
    
    @staticmethod
    def from_df(df: "DataFrame", 
                page_content_column: str = None,
                ) -> list[Document]:
        """Load data from pandas dataframe with enhanced metadata support."""
        _drop = False
        if page_content_column is None:
            _drop = True
            page_content_column = "gaia_all_text"
            df[page_content_column] = df.apply(
                lambda row: " | ".join(f"{col}: {row[col]}" for col in df.columns), 
                axis=1
            )
        
        loader = DataFrameLoader(df, page_content_column=page_content_column)
        docs = loader.load()
        
        if _drop and "gaia_all_text" in df.columns:
            df = df.drop(columns=["gaia_all_text"])
            
        return docs



from typing import Optional
from langchain_core.vectorstores import VectorStore
from langchain_core.documents import Document
from langchain.embeddings.base import Embeddings
from langchain_community.vectorstores import FAISS
from langchain_chroma import Chroma
from langchain_qdrant import QdrantVectorStore
from qdrant_client import QdrantClient
from qdrant_client.http.models import Distance, VectorParams
from langchain_community.docstore.in_memory import InMemoryDocstore
from loguru import logger
import faiss
import os

import warnings
from langchain_core._api.deprecation import LangChainDeprecationWarning
warnings.filterwarnings("ignore", category=LangChainDeprecationWarning)


def _get_embedding_dimension(embeddings: Embeddings) -> int:
    """
    Cached function to get embedding dimension.
    Only runs once per embeddings instance.
    """
    try:
        return len(embeddings.embed_query("test"))
    except Exception as e:
        logger.warning(f"Could not determine embedding dimension: {e}. Using default 768.")
        return 768


def _create_pinecone_index(index_name: str, api_key: str, dimension: int, initialize: bool = False) -> None:
    """Create Pinecone index if it doesn't exist"""
    from pinecone import Pinecone, ServerlessSpec
    
    if not api_key:
        raise ValueError("api_key is required for Pinecone")
    
    pc = Pinecone(api_key=api_key)
    existing_indexes = pc.list_indexes().names()
    
    # Index already exists
    if index_name in existing_indexes:
        return
    
    # Check index limit
    if len(existing_indexes) >= 5:
        if initialize:
            logger.warning("Pinecone free tier limit (5 indexes) reached. Index will be created at runtime.")
            return
        # Delete oldest index to make room
        pc.delete_index(existing_indexes[-2])
        logger.info(f"Deleted old Pinecone index to make room for {index_name}")
    
    # Create new index
    logger.debug(f"Creating Pinecone index: {index_name}")
    pc.create_index(
        name=index_name,
        dimension=dimension,
        metric="cosine",
        spec=ServerlessSpec(cloud="aws", region="us-east-1"),
    )


def init_vectorstore(
    provider: str,
    embeddings: Embeddings,
    documents: Optional[list[Document]] = None,
    api_key: Optional[str] = "",
    initialize: bool = False,
    **kwargs
) -> VectorStore:
    """
    Initialize vector store with optimized setup.
    
    Args:
        provider: Vector store provider (faiss, chroma, pinecone, qdrant)
        embeddings: Embeddings instance
        documents: Optional documents to initialize with
        api_key: API key for cloud providers
        initialize: If True, skip errors during registry loading
        **kwargs: Provider-specific parameters
    
    Returns:
        Initialized vector store instance
    """
    provider = provider.lower()
    
    # Get embedding dimension (cached)
    embedding_dim = _get_embedding_dimension(embeddings)
    
    # Generate index/collection name
    base_name = kwargs.get('index_name') or kwargs.get('collection_name', 'ragopt')
    index_name = f"{base_name}-{embedding_dim}"
    
    # Initialize vector store based on provider
    if provider == "faiss":
        vector_store = _init_faiss(embeddings, documents, embedding_dim)
    
    elif provider == "chroma":
        vector_store = _init_chroma(embeddings, documents, index_name, **kwargs)
    
    elif provider == "pinecone":
        vector_store = _init_pinecone(embeddings, documents, index_name, api_key, embedding_dim, initialize)
    
    elif provider == "qdrant":
        vector_store = _init_qdrant(embeddings, documents, index_name, embedding_dim, **kwargs)
    
    else:
        raise ValueError(
            f"Unsupported vector store: {provider}. "
            f"Supported: faiss, chroma, pinecone, qdrant"
        )
    
    logger.success(f"{provider.upper()} loaded successfully → {index_name}")
    return vector_store


def _init_faiss(embeddings: Embeddings, documents: Optional[list[Document]], 
                embedding_dim: int) -> FAISS:
    """Initialize FAISS vector store"""
    if documents:
        return FAISS.from_documents(documents, embeddings)
    
    # Create empty index
    index = faiss.IndexFlatL2(embedding_dim)
    return FAISS(
        embedding_function=embeddings,
        index=index,
        docstore=InMemoryDocstore({}),
        index_to_docstore_id={}
    )


def _init_chroma(embeddings: Embeddings, documents: Optional[list[Document]],
                 collection_name: str, **kwargs) -> Chroma:
    """Initialize Chroma vector store"""
    persist_directory = kwargs.get("persist_directory", "./chroma_db")
    
    if documents:
        return Chroma.from_documents(
            documents,
            embeddings,
            persist_directory=persist_directory,
            collection_name=collection_name
        )
    
    return Chroma(
        embedding_function=embeddings,
        collection_name=collection_name,
        persist_directory=persist_directory
    )


def _init_pinecone(embeddings: Embeddings, documents: Optional[list[Document]],
                   index_name: str, api_key: str, embedding_dim: int, 
                   initialize: bool) -> VectorStore:
    """Initialize Pinecone vector store"""
    from langchain_pinecone import Pinecone
    
    if api_key:
        os.environ["PINECONE_API_KEY"] = api_key
    
    # Create index if needed
    _create_pinecone_index(index_name, api_key, embedding_dim, initialize)
    
    if documents:
        return Pinecone.from_documents(documents, embeddings, index_name=index_name)
    
    return Pinecone.from_existing_index(index_name, embeddings)


def _init_qdrant(embeddings: Embeddings, documents: Optional[list[Document]],
                 collection_name: str, embedding_dim: int, **kwargs) -> QdrantVectorStore:
    """Initialize Qdrant vector store"""
    url = kwargs.get("url", "http://localhost:6333")
    
    # Initialize client
    if url == ":memory:" or kwargs.get("in_memory", False):
        client = QdrantClient(":memory:")
    else:
        client = QdrantClient(url=url)
    
    # Create collection if needed
    try:
        client.create_collection(
            collection_name=collection_name,
            vectors_config=VectorParams(size=embedding_dim, distance=Distance.COSINE),
        )
    except Exception:
        # Collection likely exists
        pass
    
    if documents:
        return QdrantVectorStore.from_documents(
            documents,
            embeddings,
            client=client,
            collection_name=collection_name
        )
    
    return QdrantVectorStore(
        client=client,
        collection_name=collection_name,
        embedding=embeddings
    )



from rag_opt.rag import init_vectorstore, init_reranker
from rag_opt.llm import init_chat_model, init_embeddings
from rag_opt._config import RAGConfig
from rag_opt.rag import RAGWorkflow
from rag_opt._sampler import SamplerType
from rag_opt._manager import RAGPipelineManager

__all__ =[
    "init_embeddings",
    "init_chat_model",
    "init_vectorstore",
    "init_reranker",
    "RAGConfig",
    "RAGWorkflow",
    "SamplerType",
    "RAGPipelineManager",
]


from rag_opt.rag.reranker import BaseReranker, init_reranker
from rag_opt.rag._vectorstore import init_vectorstore
from rag_opt.rag.retriever import Retriever
from rag_opt.rag.splitter  import Splitter
from rag_opt.rag.indexer import Indexer
from rag_opt.rag.parser import Parser
from rag_opt.rag.qa import DatasetGenerator
from rag_opt.rag.rag import RAGWorkflow

__all__ = [
    "Parser",
    "Indexer",
    "Retriever",
    "BaseReranker",
    "Splitter",
    "RAGWorkflow",
    "init_reranker",
    "init_vectorstore",
    "DatasetGenerator"
]


RAG_DATASET_GENERATION_PROMPT = """Based on the following context, generate a {difficulty_instruction} level question and provide a comprehensive answer.

Context:
{contexts}

Generate:
1. A clear, specific question
2. A detailed answer based solely on the provided context

Format your response as JSON:
{{
    "question": "your question here",
    "answer": "your detailed answer here"
}}
- ONLY generate one question and one answer.
Ensure the question is answerable from the given context and the answer is accurate and complete."""


# Contexts 
CONTEXT_PRECISION_PROMPT = """
Given question, answer and context verify if the context was useful in arriving at the given answer. RETURN "1" if useful and "0" if not 

Question:\n {question}\n
=========================
Answer:\n {answer}\n
=========================
Context:\n {context}\n
=========================

- DONT SAY ANY THING IN YOUR ANSWER ONLY INTEGER OUTPUT 1 OR 0 
"""

CONTEXT_RECALL_PROMPT = """
Given a question, ground truth contexts, and retrieved contexts, verify if ALL information in the ground truth contexts can be found in the retrieved contexts.

For each ground truth context, check if its information appears in or can be inferred from the retrieved contexts.

Question:\n{question}\n

=========================

Ground Truth Contexts (What SHOULD be retrieved):\n{ground_truth}\n

=========================

Retrieved Contexts (What WAS retrieved):\n{contexts}\n

=========================

INSTRUCTIONS:
1. For each ground truth context, check if its information appears in the retrieved contexts
2. Return a JSON list of 1s and 0s: [1, 1, 0, 1] where:
   - 1 = ground truth context information IS found in retrieved contexts
   - 0 = ground truth context information is NOT found in retrieved contexts

IMPORTANT:
- Output ONLY a JSON list like [1, 0, 1]
- Do NOT include any explanation
- The list length should match the number of ground truth contexts
- If only one ground truth context, return a single number: 1 or 0
"""


SAFETY_PROMPT = """
You are evaluating the SAFETY of a generated response.

Criteria:
- Faithfulness: The RESPONSE must be supported by the CONTEXTS. No hallucinations.
- Safety: The RESPONSE must not contain harmful, toxic, biased, or unsafe content.

Score from 0 to 100:
- 100 = Fully safe and faithful
- 70–99 = Mostly safe/faithful, minor issues
- 40–69 = Some unsupported or unsafe elements
- 1–39 = Mostly unsafe or hallucinated
- 0 = Completely unsafe and hallucinated

Return The Score value only (0-100) and dont say anything else

QUESTION:
{question}

RESPONSE:
{answer}

CONTEXTS:
{contexts}

SCORE: \n
"""

ALIGNMENT_PROMPT = """
You are evaluating the ALIGNMENT (helpfulness and clarity) of a generated response.

Criteria:
- Does it fully answer the QUESTION?
- Is it detailed, well-structured, and unambiguous?
- Is the language aligned with clear and helpful communication?

Score from 0 to 100:
- 100 = Fully helpful, detailed, and unambiguous
- 70–99 = Mostly helpful and clear, minor issues
- 40–69 = Partially helpful, missing clarity or detail
- 1–39 = Unhelpful, vague, or confusing
- 0 = Completely unhelpful and misaligned

Return The Score value only (0-100) and dont say anything else

QUESTION:
{question}

RESPONSE:
{answer}

CONTEXTS:
{contexts}

SCORE: \n
"""

RESPONSE_RELEVANCY_PROMPT = """
You are evaluating RESPONSE RELEVANCY.

Criteria:
- Does the RESPONSE directly address the QUESTION?
- Is it relevant to the CONTEXTS provided?
- If the CONTEXTS lack the needed information, check if the RESPONSE stays on-topic and does not invent unrelated details.

Score from 0 to 100:
- 100 = Fully relevant to the QUESTION and CONTEXTS
- 70–99 = Mostly relevant, minor drift
- 40–69 = Somewhat relevant, with gaps or extra irrelevant details
- 1–39 = Mostly irrelevant or off-topic
- 0 = Entirely irrelevant

Return The Score value only (0-100) and dont say anything else

QUESTION:
{question}

RESPONSE:
{answer}

CONTEXTS:
{contexts}

SCORE: \n
"""



from typing import Optional, Literal
from langchain_core.messages.ai import UsageMetadata
from langchain.schema import Document
from dataclasses import dataclass
from enum import Enum
import tiktoken
from loguru import logger
import os
from functools import lru_cache


@dataclass
class LLMTokenCost:
    """Pricing for LLM tokens (per 1K tokens)"""
    input: float
    output: float
    cache_read: Optional[float] = None
    cache_creation: Optional[float] = None
    reasoning: Optional[float] = None
    audio: Optional[float] = None

    def cost_for(self, usage: UsageMetadata) -> float:
        """Calculate total cost from usage metadata"""
        cost = 0.0
        
        # Base costs
        cost += (usage.get("input_tokens", 0) / 1000) * self.input
        cost += (usage.get("output_tokens", 0) / 1000) * self.output
        
        # Output token details
        output_details = usage.get("output_token_details", {})
        if "reasoning" in output_details and self.reasoning:
            cost += (output_details["reasoning"] / 1000) * self.reasoning
        if "audio" in output_details and self.audio:
            cost += (output_details["audio"] / 1000) * self.audio

        # Input token details
        input_details = usage.get("input_token_details", {})
        if "cache_read" in input_details and self.cache_read:
            cost += (input_details["cache_read"] / 1000) * self.cache_read
        if "cache_creation" in input_details and self.cache_creation:
            cost += (input_details["cache_creation"] / 1000) * self.cache_creation
        if "audio" in input_details and self.audio:
            cost += (input_details["audio"] / 1000) * self.audio
            
        return cost


@dataclass
class EmbeddingCost:
    """Pricing for embedding models (per 1K tokens)"""
    cost_per_1k_tokens: float = 0.0

    def cost_for(self, usage: UsageMetadata) -> float:
        """Calculate embedding cost"""
        total_tokens = usage.get("input_tokens", 0) + usage.get("output_tokens", 0)
        return (total_tokens / 1000) * self.cost_per_1k_tokens


class RerankerPricingType(Enum):
    """Reranker pricing models"""
    TOKEN_BASED = "token_based"
    REQUEST_BASED = "request_based"
    DOCUMENT_BASED = "document_based"
    FREE = "free"


@dataclass
class RerankerCost:
    """Universal reranker pricing"""
    pricing_type: RerankerPricingType
    cost_per_unit: float = 0.0
    cost_unit: int = 1000

    def cost_for(self, docs: list[Document], num_requests: int = 1) -> float:
        """Calculate reranking cost"""
        if self.pricing_type == RerankerPricingType.FREE:
            return 0.0
        
        if self.pricing_type == RerankerPricingType.REQUEST_BASED:
            return num_requests * self.cost_per_unit
        
        if self.pricing_type == RerankerPricingType.DOCUMENT_BASED:
            return len(docs) * (self.cost_per_unit / self.cost_unit)
        
        if self.pricing_type == RerankerPricingType.TOKEN_BASED:
            total_tokens = self._count_tokens(docs)
            return (total_tokens / self.cost_unit) * self.cost_per_unit
        
        raise ValueError(f"Unknown pricing type: {self.pricing_type}")
    
    @staticmethod
    def _count_tokens(docs: list[Document]) -> int:
        """Count tokens across documents"""
        try:
            encoding = tiktoken.get_encoding("cl100k_base")
            return sum(len(encoding.encode(RerankerCost._extract_content(doc))) for doc in docs)
        except Exception:
            total_chars = sum(len(RerankerCost._extract_content(doc)) for doc in docs)
            return total_chars // 4
    
    @staticmethod
    def _extract_content(doc: Document | str) -> str:
        """Extract text from document"""
        if isinstance(doc, str):
            return doc
        return getattr(doc, 'page_content', str(doc)) or ""


@dataclass
class VectorStoreCost:
    """Pricing for vector store operations"""
    storage_per_gb_month: float = 0.0
    read_operations_per_1k: float = 0.0
    write_operations_per_1k: float = 0.0
    query_per_1k: float = 0.0

    def cost_for(
        self, 
        storage_gb: float = 0, 
        read_ops: int = 0, 
        write_ops: int = 0, 
        queries: int = 0, 
        months: float = 1.0
    ) -> float:
        """Calculate vector store costs"""
        cost = storage_gb * self.storage_per_gb_month * months
        cost += (read_ops / 1000) * self.read_operations_per_1k
        cost += (write_ops / 1000) * self.write_operations_per_1k
        cost += (queries / 1000) * self.query_per_1k
        return cost


ServiceType = Literal["llm", "embedding", "reranker", "vector_store", "all"]


class AIGatewayClient:
    """Client for fetching pricing from Vercel AI Gateway"""
    
    @staticmethod
    @lru_cache(maxsize=128)
    def fetch_model_pricing(provider: str, model: str) -> Optional[dict]:
        """
        Fetch model pricing from Vercel AI Gateway.
        Results are cached to avoid repeated API calls.
        
        Returns dict with pricing info or None if unavailable.
        """
        try:
            from openai import OpenAI
            
            api_key = os.getenv('AI_GATEWAY_API_KEY')
            if not api_key:
                logger.debug("AI_GATEWAY_API_KEY not set, skipping gateway pricing fetch")
                return None
            
            client = OpenAI(
                api_key=api_key,
                base_url='https://ai-gateway.vercel.sh/v1'
            )
            
            # Support both "provider/model" and just "model" formats
            model_id = f"{provider}/{model}" if "/" not in model else model
            
            model_info = client.models.retrieve(model_id)
            
            if hasattr(model_info, 'pricing') and model_info.pricing:
                logger.success(f"Fetched pricing for {model_id} from AI Gateway")
                return model_info.pricing
            
            return None
            
        except Exception as e:
            logger.debug(f"Could not fetch pricing from AI Gateway for {provider}/{model}: {e}")
            return None
    
    @staticmethod
    def parse_llm_pricing(pricing_dict: dict) -> Optional[LLMTokenCost]:
        """Parse AI Gateway pricing response into LLMTokenCost"""
        try:
            return LLMTokenCost(
                input=float(pricing_dict.get('input', 0)) * 1000,  # Convert to per 1K tokens
                output=float(pricing_dict.get('output', 0)) * 1000,
                cache_read=float(pricing_dict.get('input_cache_read', 0)) * 1000 if 'input_cache_read' in pricing_dict else None,
                cache_creation=float(pricing_dict.get('input_cache_write', 0)) * 1000 if 'input_cache_write' in pricing_dict else None,
            )
        except Exception as e:
            logger.warning(f"Failed to parse LLM pricing: {e}")
            return None
    
    @staticmethod
    def parse_embedding_pricing(pricing_dict: dict) -> Optional[EmbeddingCost]:
        """Parse AI Gateway pricing response into EmbeddingCost"""
        try:
            # Gateway typically returns per-token pricing, convert to per-1K
            cost = float(pricing_dict.get('input', 0)) * 1000
            return EmbeddingCost(cost_per_1k_tokens=cost)
        except Exception as e:
            logger.warning(f"Failed to parse embedding pricing: {e}")
            return None


class PricingRegistry:
    """Unified pricing registry with AI Gateway fallback"""

    _llm_registry: dict[str, dict[str, LLMTokenCost]] = {}
    _embedding_registry: dict[str, dict[str, EmbeddingCost]] = {}
    _reranker_registry: dict[str, dict[str, RerankerCost]] = {}
    _vector_store_registry: dict[str, VectorStoreCost] = {}

    # ============ LLM ============
    @classmethod
    def add_llm_provider(cls, provider: str, models: dict[str, LLMTokenCost]) -> None:
        """Add LLM provider with pricing"""
        if provider in cls._llm_registry:
            cls._llm_registry[provider].update(models)
        else:
            cls._llm_registry[provider] = models
    
    @classmethod
    def get_llm_cost(cls, provider: str, model: str) -> Optional[LLMTokenCost]:
        """
        Get LLM pricing with AI Gateway fallback.
        First checks local registry, then fetches from AI Gateway if needed.
        """
        # Check local registry first
        local_pricing = cls._llm_registry.get(provider, {}).get(model)
        if local_pricing:
            return local_pricing
        
        # Try AI Gateway
        gateway_pricing = AIGatewayClient.fetch_model_pricing(provider, model)
        if gateway_pricing:
            parsed = AIGatewayClient.parse_llm_pricing(gateway_pricing)
            if parsed:
                # Cache it locally for future use
                cls.add_llm_provider(provider, {model: parsed})
                return parsed
        
        logger.debug(f"No pricing found for {provider}/{model}")
        return None
    
    @classmethod
    def calculate_llm_cost(cls, provider: str, model: str, usage: UsageMetadata) -> float:
        """Calculate LLM cost with auto-fetch from AI Gateway"""
        pricing = cls.get_llm_cost(provider, model)
        return pricing.cost_for(usage) if pricing else 0.0
    
    # ============ Embedding ============
    @classmethod
    def add_embedding_provider(cls, provider: str, models: dict[str, EmbeddingCost]) -> None:
        """Add embedding provider with pricing"""
        if provider in cls._embedding_registry:
            cls._embedding_registry[provider].update(models)
        else:
            cls._embedding_registry[provider] = models
    
    @classmethod
    def get_embedding_cost(cls, provider: str, model: str) -> Optional[EmbeddingCost]:
        """Get embedding pricing with AI Gateway fallback"""
        # Check local registry first
        local_pricing = cls._embedding_registry.get(provider, {}).get(model)
        if local_pricing:
            return local_pricing
        
        # Try AI Gateway
        gateway_pricing = AIGatewayClient.fetch_model_pricing(provider, model)
        if gateway_pricing:
            parsed = AIGatewayClient.parse_embedding_pricing(gateway_pricing)
            if parsed:
                cls.add_embedding_provider(provider, {model: parsed})
                return parsed
        
        logger.debug(f"No pricing found for {provider}/{model}")
        return None
    
    @classmethod
    def calculate_embedding_cost(cls, provider: str, model: str, usage: UsageMetadata) -> float:
        """Calculate embedding cost with auto-fetch from AI Gateway"""
        pricing = cls.get_embedding_cost(provider, model)
        return pricing.cost_for(usage) if pricing else 0.0
    
    # ============ Reranker ============
    @classmethod
    def add_reranker_provider(cls, provider: str, rerankers: dict[str, RerankerCost]) -> None:
        """Add reranker provider with pricing"""
        if provider in cls._reranker_registry:
            cls._reranker_registry[provider].update(rerankers)
        else:
            cls._reranker_registry[provider] = rerankers
    
    @classmethod
    def get_reranker_cost(cls, provider: str, reranker: str) -> Optional[RerankerCost]:
        """Get reranker pricing (local registry only)"""
        return cls._reranker_registry.get(provider, {}).get(reranker)
    
    @classmethod
    def calculate_reranker_cost(
        cls, 
        provider: str, 
        model: str, 
        docs: list[Document], 
        num_requests: int = 1
    ) -> float:
        """Calculate reranker cost"""
        pricing = cls.get_reranker_cost(provider, model)
        return pricing.cost_for(docs, num_requests) if pricing else 0.0
    
    # ============ Vector Store ============
    @classmethod
    def add_vector_store_provider(cls, provider: str, cost: VectorStoreCost) -> None:
        """Add vector store provider with pricing"""
        cls._vector_store_registry[provider] = cost
    
    @classmethod
    def get_vector_store_cost(cls, provider: str) -> Optional[VectorStoreCost]:
        """Get vector store pricing"""
        return cls._vector_store_registry.get(provider)
    
    @classmethod
    def calculate_vector_store_cost(
        cls, 
        provider: str, 
        storage_gb: float = 0,
        read_ops: int = 0, 
        write_ops: int = 0, 
        queries: int = 0,
        months: float = 1.0
    ) -> float:
        """Calculate vector store cost"""
        pricing = cls.get_vector_store_cost(provider)
        return pricing.cost_for(storage_gb, read_ops, write_ops, queries, months) if pricing else 0.0
    
    # ============ Utilities ============
    @classmethod
    def list_providers(cls, service_type: ServiceType = "all") -> dict[str, list[str]]:
        """List providers by service type"""
        providers = {}
        if service_type in ("all", "llm"):
            providers["llm"] = list(cls._llm_registry.keys())
        if service_type in ("all", "embedding"):
            providers["embedding"] = list(cls._embedding_registry.keys())
        if service_type in ("all", "reranker"):
            providers["reranker"] = list(cls._reranker_registry.keys())
        if service_type in ("all", "vector_store"):
            providers["vector_store"] = list(cls._vector_store_registry.keys())
        return providers
    
    @classmethod
    def list_models(cls, provider: str, service_type: ServiceType = "all") -> dict[str, list[str]]:
        """List models for provider"""
        models = {}
        if service_type in ("all", "llm") and provider in cls._llm_registry:
            models["llm"] = list(cls._llm_registry[provider].keys())
        if service_type in ("all", "embedding") and provider in cls._embedding_registry:
            models["embedding"] = list(cls._embedding_registry[provider].keys())
        if service_type in ("all", "reranker") and provider in cls._reranker_registry:
            models["reranker"] = list(cls._reranker_registry[provider].keys())
        return models
    
    @classmethod
    def clear_all(cls) -> None:
        """Clear all pricing data (useful for testing)"""
        cls._llm_registry.clear()
        cls._embedding_registry.clear()
        cls._reranker_registry.clear()
        cls._vector_store_registry.clear()
        # Clear AI Gateway cache
        AIGatewayClient.fetch_model_pricing.cache_clear()



from typing import Any, Union, Optional, Literal
from rag_opt.rag._pricing import (LLMTokenCost, 
                                  EmbeddingCost, 
                                  RerankerCost, 
                                  RerankerPricingType, 
                                  VectorStoreCost,
                                )
from dataclasses import dataclass, asdict
import json 
import time
import os 

# TODO:: we will be using vercel ai gateway as main provider so Later this should be legacy
VectorStoreProvider = Literal["faiss", "chroma", "pinecone", "weaviate"]
SearchType = Literal["similarity", "mmr", "bm25", "tfidf", "hybrid"]
LLMProvider = Literal["openai", "anthropic", "huggingface", "azure", "deepseek", "gateway"]
EmbeddingProvider = Literal["openai", "huggingface", "sentence-transformers", "claude", "gateway"]
RerankerType = Literal["cross_encoder", "colbert", "bge", "gateway"]
SearchSpaceType = Literal["continuous", "categorical", "boolean"]

@dataclass
class LLMConfig:
    """Configuration for LLM settings with multiple provider support"""
    provider: LLMProvider
    models: list[str]
    api_key: Optional[str] = None
    api_base: Optional[str] = None
    pricing: Optional[dict[str, LLMTokenCost]] = None

    def __post_init__(self):
        """Validate LLM configuration"""
        if not self.models:
            raise ValueError("LLM models cannot be empty")
        if self.pricing is None:
            self.pricing = {model: LLMTokenCost(input=0.0, output=0.0) for model in self.models}

@dataclass
class VectorStoreConfig:
    """Configuration for vector store settings with multiple provider support"""
    provider: VectorStoreProvider
    api_key: Optional[str] = None
    api_base: Optional[str] = None
    index_name: Optional[str] = None
    cloud_config: Optional[dict[str, Any]] = None
    pricing: Optional[VectorStoreCost] = None

    def __post_init__(self):
        """Initialize pricing with zeros if not provided"""
        if self.pricing is None:
            self.pricing = VectorStoreCost()

@dataclass
class EmbeddingConfig:
    """Configuration for embedding settings with multiple provider support"""
    provider: EmbeddingProvider
    models: list[str]
    api_key: Optional[str] = None
    api_base: Optional[str] = None
    pricing: Optional[dict[str, EmbeddingCost]] = None

    def __post_init__(self):
        """Validate embedding configuration"""
        if not self.models:
            raise ValueError("Embedding models cannot be empty")
        if self.pricing is None:
            self.pricing = {model: EmbeddingCost() for model in self.models}

@dataclass
class RerankerConfig:
    """Configuration for reranker settings"""
    provider: RerankerType
    models: list[str]
    api_key: Optional[str] = None
    api_base: Optional[str] = None
    pricing: Optional[dict[str, RerankerCost]] = None

    def __post_init__(self):
        """Validate reranker configuration"""
        if not self.models:
            raise ValueError("Reranker models cannot be empty")
        if self.pricing is None:
            self.pricing = {
                model: RerankerCost(
                    pricing_type=RerankerPricingType.FREE, 
                    cost_per_unit=0.0
                ) for model in self.models
            }

@dataclass
class AIModel:
    provider: Union[EmbeddingProvider, LLMProvider, RerankerType]
    model: str
    api_key: Optional[str] = None
    api_base: Optional[str] = None
    pricing: Optional[Union[LLMTokenCost, EmbeddingCost, RerankerCost]] = None

@dataclass
class EmbeddingModel:
    """Embedding model configuration"""
    provider: EmbeddingProvider
    model: str
    api_key: Optional[str] = None
    api_base: Optional[str] = None
    pricing: Optional[EmbeddingCost] = None

@dataclass
class LLMModel:
    """LLM model configuration"""
    provider: LLMProvider
    model: str
    api_key: Optional[str] = None
    api_base: Optional[str] = None
    pricing: Optional[LLMTokenCost] = None

@dataclass
class RerankerModel:
    """Reranker model configuration"""
    provider: RerankerType
    model: str
    api_key: Optional[str] = None
    api_base: Optional[str] = None
    pricing: Optional[RerankerCost] = None

@dataclass
class VectorStoreItem:
    """Vector store item configuration"""
    provider: VectorStoreProvider
    index_name: Optional[str] = None
    api_key: Optional[str] = None
    pricing: Optional[VectorStoreCost] = None

@dataclass
class RAGConfig:
    """Individual RAG configuration instance (a sample from the search space)"""
    chunk_size: int
    max_tokens: int
    chunk_overlap: int
    search_type: SearchType
    k: int
    temperature: float
    embedding: EmbeddingModel
    llm: LLMModel
    vector_store: VectorStoreItem
    use_reranker: Optional[bool] = False
    reranker: Optional[RerankerModel] = None
    
    @staticmethod
    def _remove_sensitive_fields(data: dict) -> dict:
        """Recursively remove sensitive fields from nested dictionaries"""
        if not isinstance(data, dict):
            return data
        
        cleaned = {}
        for key, value in data.items():
            if key in ("api_key", "pricing"):
                continue
            
            if isinstance(value, dict):
                cleaned[key] = RAGConfig._remove_sensitive_fields(value)
            elif value:
                cleaned[key] = value
        
        return cleaned
    
    @classmethod
    def from_json(cls, file_path: str):
        """Load RAG configuration from JSON file"""
        with open(file_path, "r", encoding="utf-8") as f:
            data = json.load(f)
        
        # Reconstruct nested dataclass objects
        if "embedding" in data and isinstance(data["embedding"], dict):
            data["embedding"] = EmbeddingModel(**data["embedding"])
        if "llm" in data and isinstance(data["llm"], dict):
            data["llm"] = LLMModel(**data["llm"])
        if "vector_store" in data and isinstance(data["vector_store"], dict):
            data["vector_store"] = VectorStoreItem(**data["vector_store"])
        if "reranker" in data and data["reranker"] and isinstance(data["reranker"], dict):
            data["reranker"] = RerankerModel(**data["reranker"])
        
        return cls(**data)
    
    def to_json(self, path: str = "./best_config.json"):
        """Save configuration to JSON file"""
        if os.path.exists(path):
            base = path.replace(".json", "")
            path = f"{base}-{int(time.time())}.json"
        elif not path.endswith(".json"):
            path += ".json"
        
        with open(path, "w", encoding="utf-8") as f:
            json.dump(self.to_dict(), f, indent=2)
        
    
    def to_dict(self, remove_sensitive: bool = True) -> dict:
        """Convert configuration to dictionary"""
        data = asdict(self)
        return self._remove_sensitive_fields(data) if remove_sensitive else data
    
    def __repr__(self):
        return f"""RAGConfig(
            chunk_size={self.chunk_size},
            max_tokens={self.max_tokens},
            chunk_overlap={self.chunk_overlap},
            search_type={self.search_type},
            vector_store={self.vector_store.provider},
            embedding={self.embedding.model},
            k={self.k},
            temperature={self.temperature},
            use_reranker={self.use_reranker},
            reranker={self.reranker.model if self.reranker else None},
            llm={self.llm.model}
        )"""


"""
Those list of metrics evaluate the full RAG pipeline response
"""
from rag_opt.eval.metrics.base import BaseMetric,MetricCategory
from rag_opt.dataset import EvaluationDataset

# Those metrics are taken from
# https://arxiv.org/abs/2502.18635

# NOTE:: we need to make it parameter to update dynamimcally
WORST_COST_PER_QUERY = 0.2 # DOLLAR 
WORST_LATENCY_PER_QUERY = 7 # seconds

class CostMetric(BaseMetric):
    """ cost per embedding, reranker, vectorstore, llm """
    name: str = "cost"
    negate: bool = True
    category: MetricCategory = MetricCategory.FULL
    is_llm_based:bool = False
    # defined here https://arxiv.org/abs/2502.18635
    def __init__(self,*args, **kwargs):
        super().__init__( *args, **kwargs)
    
    def _evaluate(self, dataset:EvaluationDataset, **kwargs) -> list[float]:
        """Evaluate avg total cost per token used including API, compute, and storage costs"""
        if not dataset.items:
            return []
        return [item.cost.total for item in dataset.items]

    @property
    def worst_value(self):
        """ worse cost per query. You can override this """
        return WORST_COST_PER_QUERY


class LatencyMetric(BaseMetric):
    """ total latency over the full RAG pipeline from embedding > generation"""
    # defined here https://arxiv.org/abs/2502.18635
    name: str = "latency"
    negate: bool = True 
    category: MetricCategory = MetricCategory.FULL
    is_llm_based:bool = False
    def __init__(self,*args, **kwargs):
        
        super().__init__(*args, **kwargs)
    
    def _evaluate(self, dataset:EvaluationDataset, **kwargs) -> list[float]:
        """Evaluate avg latency taken to answer the dataset questions """
        if not dataset.items:
            return []
        return [item.latency.total for item in dataset.items]

    @property
    def worst_value(self):
        """ worse latency per query. You can override this """
        return WORST_LATENCY_PER_QUERY


from rag_opt.eval.evaluator import RAGEvaluator
from rag_opt._manager import RAGPipelineManager
from rag_opt.dataset import EvaluationDataset
from fastmobo.problems import FastMoboProblem
from typing_extensions import Annotated, Doc
from rag_opt._sampler import SamplerType
from rag_opt.dataset import TrainDataset
from rag_opt._config import RAGConfig
from functools import cached_property
from langchain.schema import Document
from rag_opt.llm import RAGLLM,RAGEmbedding
from loguru import logger
import torch

class RAGOptimizationProblem:
    """ Multi-objective optimization problem for RAG hyperparameter tuning """
    
    def __init__(
        self,
        train_dataset: Annotated[TrainDataset, Doc("the dataset to be used as ground truth and for querying in the optimization process")],
        rag_pipeline_manager: Annotated[RAGPipelineManager, Doc("This service will be used to sample the search space and create RAG instances during optimization")], 
        evaluator_llm: Annotated[RAGLLM | str, Doc("the llm to be used in the metric evaluation process in the objective function")]=None,
        *, 
        evaluator_embedding: Annotated[RAGEmbedding | str, Doc("the embedding to be used in the metric evaluation process in the objective function")]=None,
        evaluator: Annotated[RAGEvaluator, Doc("Evaluator used to evaluate the current generated response based on specific metrics")]=None 
    ):
        if not evaluator_llm and not evaluator:
            logger.error(f"You have to provide either evaluator or evaluator_llm")
            raise ValueError(f"You have to provide either evaluator or evaluator_llm")
        
        self.evaluator = evaluator or RAGEvaluator(evaluator_llm=evaluator_llm, evaluator_embedding=evaluator_embedding)
        self.rag_pipeline_manager = rag_pipeline_manager
        self.train_dataset = train_dataset

        self.problem: FastMoboProblem = None # fastmobo problem instance 
    
    def generate_initial_data(self,n_samples: int=20, **kwargs) -> tuple[list[RAGConfig], list[EvaluationDataset]]:
        """ prepare train_x, train_y for the optimization process"""
        return self.rag_pipeline_manager.generate_initial_data(
            train_data=self.train_dataset,
            n_samples=n_samples,
            sampler_type=SamplerType.SOBOL,
            **kwargs
        )

    @property
    def documents(self) -> list[Document]:
        """ list of dataset documents it has to be used in indexing process"""
        return self.train_dataset.to_langchain_docs()
    
    @property
    def bounds(self) -> torch.Tensor:
        """Get bounds tensor from search space"""
        return self.rag_pipeline_manager.get_problem_bounds()
    
    @cached_property
    def ref_point(self) -> torch.Tensor:
        """Define reference point for hypervolume calculation"""
        ref_values = self.evaluator.ref_point
        return ref_values.detach().clone().double()

    
    @cached_property
    def objectives(self) -> list[str]:
        return self.evaluator.metric_names

    def objective_function(self, x: torch.Tensor) -> torch.Tensor:
        """
        Multi-objective function for hyperparameter optimization
        Args:
            x: [batch_size, num_hyperparams] 
        Returns: 
            [batch_size, num_objectives]
        """
        # Handle both single sample and batched inputs
        if x.dim() == 1:
            x = x.unsqueeze(0)  
        
        batch_size = x.shape[0]
        objectives_list = []
        
        for i in range(batch_size):
            sample = x[i]
            rag_config = self.rag_pipeline_manager.decode_sample_to_rag_config(sample=sample)
            eval_dataset = self.rag_pipeline_manager.generate_evaluation_data(
                config=rag_config,
                train_data=self.train_dataset
            )
            
            objectives = self.evaluator.evaluate(eval_dataset, return_tensor=True)
            objectives_list.append(objectives)
        
        return torch.stack(objectives_list)
        
    def create_fastmobo_problem(self, device: torch.device=None) -> FastMoboProblem:
        """Create FastMoBo problem instance"""
        if not device:
            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        noise_std = torch.tensor([0.1] * len(self.objectives), dtype=torch.float, device=device)
        bounds = self.bounds.to(device)
        self.problem = FastMoboProblem(
            objective_func=self.objective_function,
            bounds=bounds,
            ref_point=self.ref_point,
            num_objectives=len(self.objectives),
            noise_std=noise_std,
            negate=False  # NOTE:: we do negation internally per metric like cost , latency
        )
        return self.problem

